matrix Model;
matrix View;
matrix Projection;
matrix ViewProjection;
matrix LightViewProjection;
float4 LightPos;
float4 CamPos;
bool Textured;
bool IsShadowAble;
bool NormalMapped;

bool ApplySSAO;
bool ApplySM;
bool ApplyNM;

float4 Ka;
float4 Kd;
float4 Ks;
float Ns;

float2 screenResolution;

Texture2D txDiffuse;
Texture2D txLinearDepthMap;
Texture2D txShadowMap;
Texture2D txNormalMap;
Texture2D txRandomVec;
Texture2D txSSAOMap;
Texture2D txSSAOMapBlurred;
Texture1D txGlowPos;



SamplerState samLinear
{
	Filter = MIN_MAG_MIP_LINEAR;
	AddressU = Wrap;
	AddressV = Wrap;
};

SamplerState samShadowMap
{
	Filter = MIN_MAG_MIP_LINEAR;
	AddressU = Clamp;
	AddressV = Clamp;
};

SamplerState samPoint
{
	Filter = MIN_MAG_MIP_POINT;
	AddressU = Wrap;
	AddressV = Wrap;
};


struct VS_INPUT
{
	float4 Pos : POSITION;
	float4 Norm : NORMAL;
	float4 Tex : TEXCOORD;
	float4 Tan : TANGENT;
	float4 BiNor : BINORMAL;
};

struct PS_INPUT
{
	float4 Pos		: SV_POSITION;
	float3 Norm		: NORMAL0;
	float2 Tex		: TEXCOORD0;
	float3 LightVec : TEXCOORD1;
	float3 ViewVec	: TEXCOORD2;
	
	// ShadowMapping extension
	float4 PosShadowMap : TEXCOORD3;
	float LightDistance : TEXCOORD4;
	
	// SSAO extension
	float4 PosEyeSpace	: TEXCOORD5;
};


//
// Vertex Shader
//
PS_INPUT VS( VS_INPUT input )
{
    PS_INPUT output = (PS_INPUT)0;
 
	float4 PosWorld = mul( input.Pos, Model );
        
    // Applying Transformations and preprocessing for PerPixel-Lighting in the PS later
    output.LightVec = normalize((float3)LightPos - (float3)PosWorld);
    output.ViewVec = normalize(CamPos - PosWorld);
    output.Pos = mul( PosWorld, ViewProjection );
	output.Norm = normalize(mul( input.Norm, (float3x3)Model ));
	output.Tex = input.Tex;
	
	// NormalMapping extension
	if (ApplyNM && NormalMapped)
	{
		float3 tangent = normalize(mul(input.Tan.xyz, (float3x3)Model));
		float3 binormal = normalize(mul(input.BiNor.xyz, (float3x3)Model)); //normalize(cross(output.Norm, tangent) * input.Tan.w);
		output.LightVec = float3(dot(tangent, output.LightVec), dot(binormal, output.LightVec), dot(output.Norm, output.LightVec));
		output.ViewVec = float3(dot(tangent, output.ViewVec), dot(binormal, output.ViewVec), dot(output.Norm, output.ViewVec));
	}
	
	// ShadowMapping extension: Get transformed vertex position in LightSource-Space and distance to light source
	output.PosShadowMap = mul( PosWorld, LightViewProjection );
	output.LightDistance = output.PosShadowMap.z / output.PosShadowMap.w;
	
	// SSAO Extension: Get VertexPos in Eye-Space
	output.PosEyeSpace = mul( PosWorld, ViewProjection );
	
    return output;
}



float2 mapViewSpace2TexCoord(float4 viewspacePos)
{
	// Get homogeneous (screen) coordinates [-1, 1] of the input vertex (divide by w component)
	float2 homogeneousPos = viewspacePos / viewspacePos.w;
	
	// Map the homogeneous coordinate to texture coordinates [0, 1] (reversing y)
	float2 texCoord = homogeneousPos / 2.0f + float2(0.5, 0.5);
	texCoord.y = 1 - texCoord.y;
	return texCoord;
}

SamplerComparisonState  FILT_PCF
{
    AddressU = Clamp;
    AddressV = Clamp;
    Filter = COMPARISON_MIN_MAG_LINEAR_MIP_POINT;
    ComparisonFunc = Less_Equal;
};


float FPC_FILTER( float2 STCoord, float fDepth)
{
	float texOffset = 1.0f / 4096;
	float filterWidth = 5;
	
	fDepth -= 0.002f;
	STCoord -= texOffset * ((filterWidth - 1)/2);
	
	float lit = 0.0f;
	
	for (int i = 0; i < filterWidth; i++)
		for (int j = 0; j < filterWidth; j++)
			lit += txShadowMap.SampleCmpLevelZero( FILT_PCF,  float2( STCoord.x + i * texOffset, STCoord.y + j * texOffset ), fDepth );
	
	return lit / pow(filterWidth, 2);
}


float isShadowed( float4 PosShadowMap)
{
	// Get homogeneous (screen) coordinates [-1, 1] of the input vertex
	PosShadowMap.xyz /= PosShadowMap.w;
	
	// Map the homogeneous coordinate to texture coordinates [0, 1] (reversing y)
	float2 shadowMapTexCoord = PosShadowMap.xy * float2(0.5f, -0.5f) + 0.5f;
    
    return FPC_FILTER(shadowMapTexCoord, PosShadowMap.z);
}


//
// Pixel Shader
//
float4 PS( PS_INPUT input ) : SV_Target
{	
	float ssaoLightFac = 1.0f;
	if (ApplySSAO)
	{	
		ssaoLightFac = txSSAOMapBlurred.Sample( samLinear, mapViewSpace2TexCoord(input.PosEyeSpace) );
		//return ssaoLightFac;
	}
	
	
	
	float3 Light = Ka.rgb * ssaoLightFac;
	float lit = 1.0f;

	if (ApplySM && IsShadowAble)
	{
		lit = isShadowed(input.PosShadowMap);
	}
	
	if (lit != 0)
	{
		float3 spec = {0, 0, 0};
		
		float3 V = normalize(input.ViewVec);
		
		float3 Norm;
		
		if (ApplyNM && NormalMapped)
			Norm = normalize(2 * txNormalMap.Sample( samLinear, input.Tex ) - 1);
		else
			Norm = normalize(input.Norm);
			
		float3 LightV = normalize(input.LightVec);
		
		float NdotL = dot(Norm, LightV);
		
		float3 R = dot(2 * Norm, LightV) * Norm - LightV;
		R = normalize(R);
		
		float VdotR = dot(V, R);
		
		if (NdotL > 0 && VdotR > 0)
			spec = Ks.rgb * pow(VdotR,Ns);
			
		Light = Kd.rgb * saturate(NdotL) + spec;
		Light *= lit;
		Light += Ka.rgb * ssaoLightFac;
	}
	
	if (Textured)
		return float4(txDiffuse.Sample( samLinear, input.Tex ) * Light, 1.0f);
	else
		return float4(Light, 1.0f);
}


struct GPS_INPUT
{
	float4 Pos		: SV_POSITION;
    float2 Tex		: TEXCOORD0;
    float4 Pos2D	: TEXCOORD1;
};



//
// Glow Vertex Shader
//
GPS_INPUT GlowVS( VS_INPUT input )
{
    GPS_INPUT output = (GPS_INPUT)0;
    
    output.Pos = mul( mul( input.Pos, Model ), ViewProjection );
	output.Pos2D = output.Pos;
	output.Tex = input.Tex;
	
    return output;
}


//
// Glow Pixel Shader
//
float4 GlowPS( GPS_INPUT input ) : SV_Target
{	
    return txDiffuse.Sample(samLinear, input.Tex) * 2.0;
}



// Technique for default Rendering
technique10 Render
{
    pass P0
    {
		SetVertexShader( CompileShader( vs_4_0, VS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PS() ) );
    }
}



//=============================================================================
// First Pass Shader for ShadowMapping
//=============================================================================

// Shadowmapping Vertex Shader
// Projection into the the View Space of the LightSource!
//-----------------------------------------------------------------------------
float4 VS_ShadowMap( VS_INPUT input ) : SV_POSITION
{
    return mul(mul(input.Pos, Model), LightViewProjection);
}


// Shadowmapping Pixel Shader
// This Shader is just used for retrieving the depth buffer into a texture in
// a first pass, so there's no use for the result color, simply return 0.
//-----------------------------------------------------------------------------
float4 PS_ShadowMap( float4 input : SV_POSITION ) : SV_Target
{
	return 0;
}

// Technique for this Shader
//-----------------------------------------------------------------------------
technique10 RenderShadowMap
{
    pass P0
    {
		SetVertexShader( CompileShader( vs_4_0, VS_ShadowMap() ) );
		SetGeometryShader( NULL );
		SetPixelShader ( CompileShader( ps_4_0, PS_ShadowMap() ) );
    }
}


//=============================================================================
// First Pass Shader for SSAO: Creating Linear Depth Map
//=============================================================================

struct SSAOPS_INPUT
{
	float4 Pos			: SV_POSITION;
	float2 Tex			: TEXCOORD0;
	float4 PosEyeSpace	: TEXCOORD1;
};	

// SSAO Vertex Shader
// Calculates Eye Space Position of the Vertex
//-----------------------------------------------------------------------------
SSAOPS_INPUT VS_LDM( VS_INPUT input )
{
	SSAOPS_INPUT output = (SSAOPS_INPUT) 0;
	
	float4 PosWorld = mul(input.Pos, Model);
	output.Pos = mul(PosWorld, ViewProjection);
	output.PosEyeSpace = mul(PosWorld, View);
	output.Tex = input.Tex;
	
	return output;
}

// SSAO Pixel Shader
// Calculates the linear eye space depth value of each Pixel (not projected!)
//-----------------------------------------------------------------------------
float4 PS_LDM( SSAOPS_INPUT input ) : SV_Target
{
	return input.PosEyeSpace;
}

// Technique for this Shader
//-----------------------------------------------------------------------------
technique10 RenderLinearDepthMap
{
    pass P0
    {
		SetVertexShader( CompileShader( vs_4_0, VS_LDM() ) );
		SetGeometryShader( NULL );
		SetPixelShader ( CompileShader( ps_4_0, PS_LDM() ) );
    }
}


//=============================================================================
// Second Pass Shader for SSAO: Render SSAO Map
//=============================================================================

// SSAO Vertex Shader
// Renders a fullscreen Quad executing SSAO calculations
//-----------------------------------------------------------------------------
SSAOPS_INPUT VS_SSAO( VS_INPUT input )
{
	SSAOPS_INPUT output = (SSAOPS_INPUT) 0;
	
	// Doesnt work as expected:
/*	float zNear = 0.1f;
	float hFovY = 45.0f;
	float aspect = 4.0f/3.0f;
	float y = tan(radians(hFovY)/2.0f)*zNear;
	float x = y * aspect;
*/	

	// So we're using this one:
	float aspect = 4.0f/3.0f;
	float planeDistance = 100.0f;
	float fac = 0.56 * planeDistance;
	float4 PosEyeSpace = input.Pos;
	PosEyeSpace.y *= fac;
	PosEyeSpace.x *= fac * aspect;
	PosEyeSpace.z = planeDistance;
	PosEyeSpace.w = 1;
	output.Pos = mul(PosEyeSpace, Projection);
		
	output.PosEyeSpace = PosEyeSpace / PosEyeSpace.z;
	output.Tex = input.Tex;
	
	return output;
}


// SSAO Pixel Shader
// ...
//-----------------------------------------------------------------------------
// Vectors of maximum length 1 describing a sphere volume
float3 ssaoSamples[16] =
{
    float3(-0.48955268, -0.7149735, 0.06780062),
	float3(0.21376659, -0.5100953, 0.35739228),
	float3(0.36881816, 0.2873843, 0.13502488),
	float3(-0.33524942, -0.22806385, 0.7403099),
	float3(-0.30810344, -0.0626009, -0.8102468),
	float3(-0.20154051, -0.33114094, 0.2554454),
	float3(-0.3694443, 0.08357073, -0.072646536),
	float3(0.6377252, -0.019399622, 0.44496638),
	float3(0.15989959, -0.52126104, 0.8333015),
	float3(0.30447853, 0.23720363, -0.20906633),
	float3(-0.22385149, 0.4570924, -0.33578417),
	float3(0.37816003, -0.7113091, -0.09377871),
	float3(-0.44385353, 0.6695935, 0.41355595),
	float3(-0.046945088, -0.6542132, -0.5632083),
	float3(-0.12570725, 0.1940638, -0.34466705),
	float3(0.32368305, -0.23416239, -0.83022475)
};

float3 randomFloat3(float x, float y)
{
	return txRandomVec.Sample( samPoint, float2(x,y) );
}

float4 PS_SSAO( SSAOPS_INPUT input ) : SV_Target
{
	float ssaoRadius = 0.3;		// TODO: Adjustable
	float attFactor = 0.5;		// TODO: Adjustable

	float4 pEye = txLinearDepthMap.Sample( samLinear, input.Tex );
	
	float4 occlusion = 0;
	for (int i=0; i<16; i++)
	{
		// Get a good seeded random number out of the given interpolated vars
		float2 pScreen = mapViewSpace2TexCoord(pEye) * 1000;
		float2 c = mapViewSpace2TexCoord(pEye*7 + float4(pScreen, 0, 1));
		float3 randomVec = normalize(randomFloat3(c.x, c.y));	
		
		// Dither the results by reflecting the Sample on the random plane normal vector
		float3 svecEye = reflect (ssaoSamples[i], randomVec);
	
		float3 sEye = pEye + svecEye * ssaoRadius;

		float2 sLDMTexCoord = mapViewSpace2TexCoord( mul(sEye, Projection) );	
		
		float sLDMDepth = txLinearDepthMap.Sample( samLinear, sLDMTexCoord ).z;
		
		// if zDiff > 0, the samplePoint is occluded by some real surface point
		float zDiff = attFactor * ( sEye.z - sLDMDepth );
		if (zDiff > 0)
		{
			occlusion += 1.0/(1.0 + zDiff*zDiff);
		}	
	}
	
	if (pEye.z < 10000.0f)	
	{	
		return 1.0 - occlusion/16;
	}
	else	
	{	
		// No SSAO applyable since no geometry
		return 1.0;
	}
}

// Technique for this Shader
//-----------------------------------------------------------------------------
technique10 RenderSSAO
{
    pass P0
    {
		SetVertexShader( CompileShader( vs_4_0, VS_SSAO() ) );
		SetGeometryShader( NULL );
		SetPixelShader ( CompileShader( ps_4_0, PS_SSAO() ) );
    }
}


//=============================================================================
// Third Pass Shader for SSAO: SSAO Bluring
//=============================================================================

// Postprocessing Pixel Shader
// Bluring
//-----------------------------------------------------------------------------
float4 PS_SSAOBlur( SSAOPS_INPUT input ) : SV_Target
{
	float4 pEye = txLinearDepthMap.Sample( samLinear, input.Tex );
	
	
	//--------------------------------------------------------
	// BLUR BYPASS!!!! COMMENT THIS TO SEE SSAO BLURRED
	
	return txSSAOMap.Sample( samLinear, input.Tex );
	//--------------------------------------------------------
	
	
	float xTexelSize = 1.0f/screenResolution.x;
	float yTexelSize = 1.0f/screenResolution.y;

	// Applay mean filtering (7x7 kernel)
	float col = 0;
	for (int i=0; i<7; i++)
		for (int j=0; j<7; j++)
		{
			float2 texCoord = input.Tex + float2(xTexelSize*(i-3), yTexelSize*(j-3));	
			float texCol = txSSAOMap.Sample( samLinear, texCoord );
			
			col += texCol;
		}
	col /= 49.0f;
	
	// Apply grayscale shifting to give the SSAO effect more presence
	//col = max(0, col-0.5);	// Shift down
	//col *= 2.0f;			// Scale to get better 0-1 contrast
		
	if (pEye.z < 10000.0f)	
	{	
		return float4(col,col,col,1);
	}
	else	
	{	
		// No SSAO blurring since no geometry
		return 1.0;
	}
}

// Technique for this Shader
//-----------------------------------------------------------------------------
technique10 BlurSSAO
{
    pass P0
    {
		SetVertexShader( CompileShader( vs_4_0, VS_SSAO() ) );
		SetGeometryShader( NULL );
		SetPixelShader ( CompileShader( ps_4_0, PS_SSAOBlur() ) );
    }
}
